/*	Winnov Videum
 *	Video for Linux Two driver
 *
 *	This program is copyright 1998, 1999
 *	by Bill Dirks
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	This is the central module for the Winnov Videum AV audio/video
 *	capture card drivers. The other modules are wnvvid.c for the 
 *	video capture driver, and wnvaud.c for the sound driver.
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif
#include <linux/config.h>
#ifndef EXPORT_SYMTAB
#define EXPORT_SYMTAB
#endif
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <linux/sched.h>
#include <linux/videodev.h>
#include <linux/version.h>
#include <asm/uaccess.h>
#include <linux/pci.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/interrupt.h>

#include "wnv.h"


/*	Variables for assigning resources via the command line
 */
/*  ISA non-PnP IO base overrides */
static int	isa0_iobase	= 0;
static int	isa1_iobase	= 0;
#ifdef MODULE_PARM
MODULE_PARM(isa0_iobase, "i");
MODULE_PARM(isa1_iobase, "i");
#endif

/*  ISA PnP IO base override */
static int	pnp0_iobase	= 0;
#ifdef MODULE_PARM
MODULE_PARM(pnp0_iobase, "i");
#endif



/*
 *	The global Videum device structure array. 
 */
struct videum_device videum[WNV_NBOARDS];

int	wnv_noirq	= 0; /* 1 = Don't use hardware interrupts */

#ifdef MODULE_PARM
MODULE_PARM(wnv_noirq, "i");
#endif


/*
 *
 *	W A V I   R E G I S T E R   A N D   S R A M   A C C E S S
 *
 */

#define BEGIN_CRITICAL_SECTION(dev)	\
        do{unsigned long flags;spin_lock_irqsave(&dev->lock,flags)
#define END_CRITICAL_SECTION(dev)	\
	spin_unlock_irqrestore(&dev->lock,flags);}while(0)

static void interrupt_handler(void *v);

/*
  wavi_readreg() and wavi_writereg() have regular and nolock versions. The
  nolock versions do not use a critical section to guarantee atomicity.
  When inside a critical section, always call the nolock functions, and
  when not inside a critical section, always call the regular functions.
  Failure to obey the rules will result in error, crash or system lockup.
*/

static __u16
wavi_readreg_nolock(struct videum_device *dev,
		    int	reg)
{
	if (dev->type == VIDEUM_TYPE_PCI)
	{
		__u16	x;
		outb(reg, LBA(dev));
		x = inw(WDATA(dev));
		return x;
	}
	if (dev->type == VIDEUM_TYPE_ISA)
	{
		return inw(dev->iobase + (reg << 10));
	}
	if (dev->type == VIDEUM_TYPE_PNPISA)
	{
		return inw(dev->iobase + (reg << 2));
	}
	return 0;
}

static __u16
wavi_writereg_nolock(struct videum_device *dev,
		     int	reg,
		     __u16	value)
{
	if (dev->type == VIDEUM_TYPE_PCI)
	{
		__u8	t = 0;
		int	dma_state = dev->dma_state;

		if (dma_state == DMA_STATE_RUNNING)
		{
			t = inb(DMAC(dev));
			outb(t & ~DMAC_HIRQEN, DMAC(dev));
		}
		outb(reg, LBA(dev));
		outw(value, WDATA(dev));
		if (dma_state == DMA_STATE_RUNNING)
		{
			outb(t, DMAC(dev));
			interrupt_handler(dev);
		}
	}
	if (dev->type == VIDEUM_TYPE_ISA)
	{
		outw(value, dev->iobase + (reg << 10));
	}
	if (dev->type == VIDEUM_TYPE_PNPISA)
	{
		outw(value, dev->iobase + (reg << 2));
	}
	return value;
}

__u16
wavi_readreg(struct videum_device *dev,
	     int	reg)
{
	if (dev->type == VIDEUM_TYPE_PCI)
	{/*	For device types that use more than one I/O call  */
		__u16	x;
		BEGIN_CRITICAL_SECTION(dev);
		x = wavi_readreg_nolock(dev, reg);
		END_CRITICAL_SECTION(dev);
		return x;
	}
	if (reg == WAVI_SPLPTR || reg == WAVI_VID1PTR || reg == WAVI_VID2PTR)
	{	/*  SPLPTR, VID1PTR and VID2PTR need to be serialized	*/
		/*  wrt block_read and block_write due to a WAVI bug	*/
		__u16	x;
		BEGIN_CRITICAL_SECTION(dev);
		x = wavi_readreg_nolock(dev, reg);
		END_CRITICAL_SECTION(dev);
		return x;
	}
	return wavi_readreg_nolock(dev, reg);
}

__u16
wavi_writereg(struct videum_device *dev,
	      int	reg,
	      __u16	value)
{
	if (dev->type == VIDEUM_TYPE_PCI)
	{/*	For device types that use more than one I/O call  */
		BEGIN_CRITICAL_SECTION(dev);
		wavi_writereg_nolock(dev, reg, value);
		END_CRITICAL_SECTION(dev);
		return value;
	}
	if (reg == WAVI_SPLPTR || reg == WAVI_VID1PTR || reg == WAVI_VID2PTR)
	{	/*  SPLPTR, VID1PTR and VID2PTR need to be serialized	*/
		/*  wrt block_read and block_write due to a WAVI bug	*/
		BEGIN_CRITICAL_SECTION(dev);
		wavi_writereg_nolock(dev, reg, value);
		END_CRITICAL_SECTION(dev);
		return value;
	}
	wavi_writereg_nolock(dev, reg, value);
	return value;
}

void
wavi_block_read(struct videum_device *dev,
		int sram_address,
		__u8 *buffer, int length)
{
	BEGIN_CRITICAL_SECTION(dev);
	if (dev->type == VIDEUM_TYPE_PCI)
	{
		wavi_writereg_nolock(dev, WAVI_HST1PTR, 
				     (__u16)(sram_address >> 4));
		outb(WAVI_HST1DAT, LBA(dev));
		insl(WDATA(dev), buffer, length >> 2);
	}
	if (dev->type == VIDEUM_TYPE_ISA)
	{
		wavi_writereg_nolock(dev, WAVI_HST1PTR, 
				     (__u16)(sram_address >> 4));
		insl(dev->iobase + (WAVI_HST1DAT << 10), buffer, length >> 2);
	}
	if (dev->type == VIDEUM_TYPE_PNPISA)
	{
		wavi_writereg_nolock(dev, WAVI_HST1PTR, 
				     (__u16)(sram_address >> 4));
		insl(dev->iobase + (WAVI_HST1DAT << 2), buffer, length >> 2);
	}
	END_CRITICAL_SECTION(dev);
}

void
wavi_block_write(struct videum_device *dev,
		 int sram_address,
		 __u8 *buffer, int length,
		 int guarantee)
{
	BEGIN_CRITICAL_SECTION(dev);
	if (dev->type == VIDEUM_TYPE_PCI)
	{
		__u8	t = 0;
		int	dma_state = dev->dma_state;

		if (guarantee && dma_state == DMA_STATE_RUNNING)
		{
			/*  Hardware bug: Can't reliably write to the card
			 *  while DMA is running.
			 */
			t = inb(DMAC(dev));
			outb(t & ~DMAC_HIRQEN, DMAC(dev));
		}
		outb(WAVI_HST1PTR, LBA(dev));
		outw((__u16)(sram_address >> 4), WDATA(dev));
		outb(WAVI_HST1DAT, LBA(dev));
		outsl(WDATA(dev), buffer, length >> 2);
		if (guarantee && dma_state == DMA_STATE_RUNNING)
		{
			outb(t, DMAC(dev));
			/*  While the HIRQEN bit is off, if a DMA trigger
			 *  or IRQ trigger occurs from the microcode, the
			 *  device will miss it and capturing will stop.
			 *  Calling interrupt_handler will restart capture
			 *  in that case, or have no effect if no event
			 *  was actually missed.
			 */
			interrupt_handler(dev);
		}
	}
	if (dev->type == VIDEUM_TYPE_ISA)
	{
		wavi_writereg_nolock(dev, WAVI_HST1PTR, 
				     (__u16)(sram_address >> 4));
		outsl(dev->iobase + (WAVI_HST1DAT << 10), buffer, length >> 2);
	}
	if (dev->type == VIDEUM_TYPE_PNPISA)
	{
		wavi_writereg_nolock(dev, WAVI_HST1PTR,
				     (__u16)(sram_address >> 4));
		outsl(dev->iobase + (WAVI_HST1DAT << 2), buffer, length >> 2);
	}
	END_CRITICAL_SECTION(dev);
}

static void
wavi_video_sampler(struct videum_device *dev, int x)
{
	__u16	t = (wavi_readreg(dev, WAVI_CTL) & ~WAVI_IMIBIT)
		    | dev->wavi_IMIBIT;
	wavi_writereg(dev, WAVI_CTL, (x) ? t | WAVI_VSEBIT : t & ~WAVI_VSEBIT);
}


/*
 *	W A V I   F U N C T I O N S
 */

static void
wavi_initialize(struct videum_device *dev)
{
	/*  Put WAVI into a sensible state and	*/
	/*  do the one-time startup things	*/
	int	t;

	t = wavi_readreg(dev, WAVI_MMA);
	switch (t & WAVI_FULLID)
	{
	case 0x02: dev->wavi_type = WINNOV_WAVI_93; break;
	case 0x12: dev->wavi_type = WINNOV_WAVI_95; break;
	case 0x22: dev->wavi_type = WINNOV_WAVI_97; break;
	default:   dev->wavi_type = 0;
		err_msg("Unknown WAVI chip\n");
		break;
	}
	if (dev->wavi_type)
		info_msg("Found WAVI-%d chip\n", dev->wavi_type);

	if (dev->wavi_type == WINNOV_WAVI_97)
	{
 		wavi_writereg(dev, WAVI_MMA, WAVI_DHRND | WAVI_MEDREN);
	}
	if (dev->type == VIDEUM_TYPE_PNPISA)
	{
		t = wavi_readreg(dev, WAVI_MMA);
		wavi_writereg(dev, WAVI_MMA, t | WAVI_BANKCS);
	}
	wavi_writereg(dev, WAVI_CTL, WAVI_VDF_YCH);
	dev->wavi_IMIBIT = 0;  /*  write only bit */
	wavi_writereg(dev, WAVI_ITR, 0);
	wavi_readreg(dev, WAVI_ITR);
	wavi_writereg(dev, WAVI_VOF, 0x8880);
	wavi_writereg(dev, WAVI_VCT, 0x4440);

	/*wavi_writereg(dev, WAVI_MBNK, 0xF0F0);*/
	/*wavi_writereg(dev, WAVI_TSTMODE, 0xFCFF);wimp*/
	if (dev->type == VIDEUM_TYPE_PNPISA)
		wavi_writereg(dev, WAVI_TSTMODE, 0xFFEE);
	if (dev->type == VIDEUM_TYPE_PCI)
		wavi_writereg(dev, WAVI_TSTMODE, 0xFFEE);

	wavi_writereg(dev, WAVI_AUDPTR, 0);
	wavi_writereg(dev, WAVI_ATTN, 0);
	dev->wavi_ATTN = 0;
	wavi_writereg(dev, WAVI_GAIN, 0);
	dev->wavi_GAIN = 0;
	wavi_writereg(dev, WAVI_FREQ, 0);
	wavi_writereg(dev, WAVI_AUDMIX1, 0);
	wavi_writereg(dev, WAVI_AUDMIX2, 0);
	dev->wavi_MIX1 = 0;
	dev->wavi_MIX2 = 0;
}


/*
 *
 *	B U S   M A S T E R   F U N C T I O N S
 *
 */



int
wnv_bm_cleanup(struct videum_device *dev)
{
	__u8	t;

	if (dev->type != VIDEUM_TYPE_PCI)
		return 0;

	BEGIN_CRITICAL_SECTION(dev);
	t = inb(DMAC(dev));
	t |=  DMAC_HIRQREQn;	/* no interrupt request */
	t |=  DMAC_HDMAEN1;	/* block dma state machine */
	outb(t, DMAC(dev));

	outb(0, BMCMD0(dev));	/* stops bus master operation */
	t = inb(BMSTATUS0(dev));
	t |=  BMSTATUS_DMA_INT;	/* clear int bit */
	t |=  BMSTATUS_ERROR;  	/* clear error bit */
	outb(t, BMSTATUS0(dev));

	outb(0, BMCMD1(dev));	/* stops bus mastering operation */
	t = inb(BMSTATUS1(dev));
	t |=  BMSTATUS_DMA_INT;	/* clear int bit */
	t |=  BMSTATUS_ERROR;  	/* clear error bit */
	outb(t, BMSTATUS1(dev));

	t = inb(DMAC(dev));
	t |=  DMAC_RSTMDMA1;	/* toggling removes MDMAEN1 */
	outb(t, DMAC(dev));
	t &= ~DMAC_RSTMDMA1;	/* ...complete the toggle */
	outb(t, DMAC(dev));

	dev->dma_state = DMA_STATE_IDLE;

	END_CRITICAL_SECTION(dev);

	return 1;
}

int
wnv_bm_resync(struct videum_device *dev)
{
	return wnv_bm_cleanup(dev);
}

int
wnv_bm_disarm(struct videum_device *dev)
{
	if (dev->dma_state == DMA_STATE_IDLE)
		return 1;
	return wnv_bm_cleanup(dev);
}

int
wnv_bm_arm(struct videum_device *dev,
	   struct scatter_node *list,
	   int swapyuyv)
{
	__u8	t;

	if (dev->type != VIDEUM_TYPE_PCI || list == NULL)
		return -1;/* error */
	wnv_bm_disarm(dev);

	BEGIN_CRITICAL_SECTION(dev);

	dev->dma_state = DMA_STATE_ARMING;

	t = inb(DMAC(dev));
	if (swapyuyv)
		t |=  DMAC_SWAP;
	else
		t &= ~DMAC_SWAP;
	outb(t, DMAC(dev));

	outl(virt_to_bus(list), BMPRDT0(dev));	/* pointer to dma list */
	outl(virt_to_bus(list), BMPRDT1(dev));	/* pointer to dma list */

	t = inb(BMCMD0(dev));
	t |=  BMCMD_RW_CNRL;	/* write mode! */
	outb(t, BMCMD0(dev));
	t |=  BMCMD_START_STOP;	/* start bus master! */
	outb(t, BMCMD0(dev));

	t = inb(BMCMD1(dev));
	t |=  BMCMD_RW_CNRL;	/* write mode! */
	outb(t, BMCMD1(dev));
	t |=  BMCMD_START_STOP;	/* start bus master! */
	outb(t, BMCMD1(dev));

	/* generate an interrupt, needed to prime the controller */
	t = inb(DMAC(dev));
	t &= ~DMAC_HIRQREQn;
	outb(t, DMAC(dev));
	t |=  DMAC_HIRQREQn;
	t &= ~DMAC_HDMAEN1;	/* let dma state machine run! */
	t |=  DMAC_HIRQEN;
	outb(t, DMAC(dev));

	END_CRITICAL_SECTION(dev);

//	if (!(inb(BMSTATUS0(dev)) & BMSTATUS_ACTIVE))
//	{
//		info_msg("Couldn't arm DMA\n");
//		dev->dma_state = DMA_STATE_IDLE;
//		return -1;
//	}
	//debug_msg("Armed DMA. State: %d\n",dev->dma_state);
	return 0;/* ok */
}

int/* 0=not done, 1=done */
wnv_bm_is_done(struct videum_device *dev)
{
	if (dev->type != VIDEUM_TYPE_PCI ||
	    dev->dma_state != DMA_STATE_RUNNING)
		return 0;
	return !(inb(BMSTATUS0(dev)) & BMSTATUS_ACTIVE);
}

int
wnv_bm_build_scatter_list(struct videum_device *dev,
			  unsigned char	   *buffer,
			  struct scatter_node **plist)
{
	struct scatter_node	*list;
	int			i, n;
	unsigned char		*a;
	int			size;

	if (dev->type != VIDEUM_TYPE_PCI || buffer == NULL)
		return 0;
	list = *plist;
	if (list == NULL)
	{/*	Assuming one page will be big enough. 4KB = 512 pieces */
		list = (struct scatter_node *)get_free_page(GFP_KERNEL);
		if (list == NULL)
			return 0;
	}

	/*  Simple algorithm will just map the buffer contiguously by pages */
	/*  Note: buffer is page-aligned */
	/*  Note2: The start of captured data may be a small offset from */
	/*         the beginning of the buffer, specified by capture_offset */

	size = dev->capture_offset + dev->capture_size;
	n = (size + PAGE_SIZE - 1) / PAGE_SIZE;
	a = buffer;
	for (i = 0; i < n; ++i)
	{
		list[i].addr = v4l2_vmalloc_to_bus(a);
		list[i].len = PAGE_SIZE;
		if (i == 0)
		{
			list[0].addr += dev->capture_offset;
			list[0].len -= dev->capture_offset;
		}
		a += PAGE_SIZE;
	}
	/* Last page might not be full */
	if (size < n * PAGE_SIZE)
		list[n - 1].len = size - (n - 1) * PAGE_SIZE;

	list[n - 1].len |= END_OF_SCATTER_LIST;

#if 0
	debug_msg("Scatter list %08lX\n", virt_to_bus(list));
	for (i = 0; i < n; ++i)
		debug_msg("List %2d: A=%08X L=%08X\n",i,
			  list[i].addr,list[i].len);
#endif
	*plist = list;
	return 1;/* ok */
}

void
wavi_benchmark(struct videum_device *dev)
{
	struct timeval	start, end;
	char		*test;
	int		i;
	u32		t;
	struct scatter_node	*list;

	if (dev->slave_rate)
		return;

	dev->slave_rate = 0;
	dev->dma_rate = 0;

	test = (char *)get_free_page(GFP_KERNEL);
	if (test == NULL)
		return;
	list = (struct scatter_node *)get_free_page(GFP_KERNEL);
	if (list == NULL)
	{
		free_page((unsigned long)test);
		return;
	}

	/*  Test slave mode SRAM readout rate. Read 1M of data in 1K */
	/*  chunks and look at elapsed time. */

	do_gettimeofday(&start);
	for (i = 0; i < 1024; ++i)
		wavi_block_read(dev, 0, test, 1024);
	do_gettimeofday(&end);
	t = (end.tv_sec - start.tv_sec) * 1000000
	  + (end.tv_usec - start.tv_usec);
	if (t < 250) t = 250;
	dev->slave_rate = 4*1048576000UL / (t / 250);

	if (dev->type == VIDEUM_TYPE_PCI && dev->irq && !wnv_noirq)
	{
		/*  Test DMA mode SRAM readout rate. Read 1M of data in */
		/*  PAGE_SIZE chunks and look at elapsed time. */

		/* This microcode will start the DMA engine */
		static char dmaucode[4] = {UI_SVS, UI_SVS, UI_RVS, UI_HLT,};
		u8 al;

		for (i = 0; i < 1048576 / PAGE_SIZE; ++i)
		{
			list[i].addr = virt_to_bus(test);
			list[i].len = PAGE_SIZE;
		}
		list[i - 1].len |= END_OF_SCATTER_LIST;

		wavi_video_sampler(dev, 0);
		wavi_block_write(dev, dev->sram_ucode, dmaucode, 4, 1);
		wavi_writereg(dev, WAVI_SPLPTR, dev->sram_ucode >> 4);

		al = inb(DMAC(dev));
		al |=  DMAC_HIRQEN;   /* Interrupts enabled */
		outb(al, DMAC(dev));
		wnv_bm_resync(dev);
		wnv_bm_arm(dev, list, 0);
		wavi_video_sampler(dev, 1);
		do_gettimeofday(&start);
		for (i = 100000/*1sec*/; i && !wnv_bm_is_done(dev); --i)
			udelay(10);/* 10 us resolution, good enough */
		do_gettimeofday(&end);
		wnv_bm_resync(dev);
		if (i == 0)
		{/*	timed out! */
			dev->dma_rate = 0;
			err_msg("Videum bus-master not working\n");
		}
		else
		{
			t = (end.tv_sec - start.tv_sec) * 1000000
			  + (end.tv_usec - start.tv_usec);
			dev->dma_rate = 4*1048576000UL / (t / 250 + 1);
		}
	}
	free_page((unsigned long)test);
	free_page((unsigned long)list);
	//debug_msg("Measured slave rate = %8d\n", dev->slave_rate);
	//debug_msg("Measured dma   rate = %8d\n", dev->dma_rate);
}


/*
 *	  2
 *	I   C   B U S   I N T E R F A C E
 *
 */

static __u16
in_ctl(struct videum_device *dev)
{
	return (wavi_readreg(dev, WAVI_CTL) & ~WAVI_IMIBIT) | dev->wavi_IMIBIT;
}
static void
out_ctl(struct videum_device *dev)
{
	wavi_writereg(dev, WAVI_CTL, dev->ctl);
}
static void
i2c_delay(void)
{
	//udelay(3);
}
static void
i2c_clock(struct videum_device *dev, int c)
{
	int	t = 1000;
	if (c) dev->ctl |= WAVI_I2CBIT; else dev->ctl &= ~WAVI_I2CBIT;
	out_ctl(dev);
	if (!c) while (--t && (in_ctl(dev) & WAVI_I2CBIT));
}
static void
i2c_data(struct videum_device *dev, int d)
{
	if (d) dev->ctl |= WAVI_IMDBIT; else dev->ctl &= ~WAVI_IMDBIT;
	out_ctl(dev);
}
/* BUG: I2C bit read routines, polarity */

static void
i2c_start(struct videum_device *dev)
{
	/* initialize the CTL register shadow */
	dev->ctl = in_ctl(dev);
	dev->ctl &= ~(WAVI_I2CBIT | WAVI_IMDBIT);
	out_ctl(dev);		i2c_delay();
	i2c_data(dev, 1);	i2c_delay();
	i2c_clock(dev, 0);	i2c_delay();
}

static int
i2c_get_ack(struct videum_device *dev)
{
	i2c_clock(dev, 1);	i2c_delay();
	i2c_data(dev, 0);	i2c_delay();
	i2c_clock(dev, 0);	i2c_delay();
				i2c_delay();
	return in_ctl(dev) & WAVI_IMDBIT;
}

static void
i2c_send_ack(struct videum_device *dev)
{
	i2c_clock(dev, 1);	i2c_delay();
	i2c_data(dev, 1);	i2c_delay();
	i2c_clock(dev, 0);	i2c_delay();
}

static void
i2c_send_nak(struct videum_device *dev)
{
	i2c_clock(dev, 1);	i2c_delay();
	i2c_data(dev, 0);	i2c_delay();
	i2c_clock(dev, 0);	i2c_delay();
}

static void
i2c_end(struct videum_device *dev)
{
	i2c_clock(dev, 1);	i2c_delay();
	i2c_data(dev, 1);	i2c_delay();
	i2c_clock(dev, 0);	i2c_delay();
	i2c_data(dev, 0);	i2c_delay();
}

static void
i2c_vpxrestart(struct videum_device *dev)
{
	i2c_clock(dev, 1);	i2c_delay();
	i2c_data(dev, 0);	i2c_delay();
	i2c_clock(dev, 0);	i2c_delay();
}

static void
i2c_out8(struct videum_device *dev,
	 int	ndata)
{
	int	i;
	
	ndata = ~ndata;
	for (i = 0; i < 8; ++i)
	{
		i2c_clock(dev, 1);		i2c_delay();
		i2c_data(dev, ndata & 128);	i2c_delay();
		i2c_clock(dev, 0);		i2c_delay();
		ndata <<= 1;
	}
}

static int
i2c_in8(struct videum_device *dev)
{
	int	i;
	int	ndata = ~0;

	dev->ctl &= ~WAVI_IMDBIT;// Don't write to h/w now
	for (i = 0; i < 8; ++i)
	{
		i2c_clock(dev, 1);	i2c_delay();
		i2c_clock(dev, 0);	i2c_delay();
		ndata = (in_ctl(dev) & 1) | (ndata << 1);
	}
	return ~ndata;
}

static int
i2c_error(struct videum_device *dev)
{
	i2c_end(dev);
	return -1;
}

void
wnv_i2c_get_ready(struct videum_device *dev)
{/*	Prepare I2C bus for operation	*/
	dev->ctl = in_ctl(dev);	/* initialize the CTL register shadow */
	i2c_end(dev);	/* idle the I2C bus */
}

int /* -1 on error */
wnv_i2c_read_reg_byte(struct videum_device *dev, 
		      int addr, int reg)
{
	int	ndata;
	
	i2c_start(dev);
	i2c_out8(dev, addr);
	if (!i2c_get_ack(dev))
		return i2c_error(dev);/* no ACK from device */
	i2c_out8(dev, reg);
	i2c_get_ack(dev); /* assume it succeeds if the first one succeeded */
	if (addr == 0x86 || addr == 0x8E)/* ITT VPX is a little different  */
		i2c_vpxrestart(dev);
	else
		i2c_end(dev);
	i2c_start(dev);
	i2c_out8(dev, addr | 1);
	if (!i2c_get_ack(dev))
		return i2c_error(dev);
	ndata = i2c_in8(dev);
	if (addr == 0x86 || addr == 0x8E)/* ITT VPX is a little different  */
		i2c_send_nak(dev);
	i2c_end(dev);
	return ndata;
}

int /* -1 on error */
wnv_i2c_write_reg_byte(struct videum_device *dev, 
		       int addr, int reg, int ndata)
{
	i2c_start(dev);
	i2c_out8(dev, addr);
	if (!i2c_get_ack(dev))
		return i2c_error(dev);/* no ACK from device */
	i2c_out8(dev, reg);
	i2c_get_ack(dev);
	i2c_out8(dev, ndata);
	i2c_get_ack(dev);
	i2c_end(dev);
	return 0;
}

int /* -1 on error */
wnv_i2c_read_reg_word(struct videum_device *dev, 
		      int addr, int reg)
{
	int	ndata;
	
	i2c_start(dev);
	i2c_out8(dev, addr);
	if (!i2c_get_ack(dev))
		return i2c_error(dev);/* no ACK from device */
	i2c_out8(dev, reg);
	i2c_get_ack(dev); /* assume it succeeds if the first one succeeded */
	if (addr == 0x86 || addr == 0x8E)/* ITT VPX is a little different  */
		i2c_vpxrestart(dev);
	else
		i2c_end(dev);
	i2c_start(dev);
	i2c_out8(dev, addr | 1);
	if (!i2c_get_ack(dev))
		return i2c_error(dev);
	ndata = i2c_in8(dev) << 8;
	i2c_send_ack(dev);
	ndata |= i2c_in8(dev);
	i2c_send_nak(dev);
	i2c_end(dev);
	return ndata;
}

int /* -1 on error */
wnv_i2c_write_reg_word(struct videum_device *dev, 
		       int addr, int reg, int ndata)
{
	i2c_start(dev);
	i2c_out8(dev, addr);
	if (!i2c_get_ack(dev))
		return i2c_error(dev);/* no ACK from device */
	i2c_out8(dev, reg);
	i2c_get_ack(dev);
	i2c_out8(dev, ndata >> 8);
	i2c_get_ack(dev);
	i2c_out8(dev, ndata & 0xFF);
	i2c_get_ack(dev);
	i2c_end(dev);
	return 0;
}

/*
 *	Read the configuration EEPROM
 */
#define EEPROM_ADDR	0xA0

static int /* 1 = OK, 0 = error */
eeprom_load(struct videum_device *dev)
{
	__u8	*p	= (__u8 *)&dev->eeprom;
	int	i;
	int	d;

	wnv_i2c_get_ready(dev);

	/*  EEPROM readability/bogosity test	*/
	if (wnv_i2c_read_reg_byte(dev, EEPROM_ADDR, 0) != EEPROM_SIG)
	{
		err_msg("Couldn't read configuration EEPROM\n");
		return 0;
	}
	for (i = 0; i < sizeof(struct EEPROM); ++i)
	{
		d = wnv_i2c_read_reg_byte(dev, EEPROM_ADDR, i);
		if (d == -1)
			break;
		p[i] = (__u8)d;
	}
        // Run checksum
	d = 0;
	for (i = 0; i < sizeof(struct EEPROM); ++i)
		d += (__u8)p[i];
	if ((__u8)d != 0xff)
	{
		err_msg("Checksum error in configuration EEPROM\n");
		return 0;
	}
	info_msg("Identified: %s\n", dev->eeprom.szProduct);
	return 1;
}



/*
 *
 *	I N T E R R U P T   R O U T I N E S
 *
 *	The interrupt is only used by the video capture driver.
 */

/*  This function runs at interrupt time, either in response to a hardware
 *  interrupt, or on each timer tick if there is no hardware interrupt.
 *  This function doesn't do anything except schedule the registered
 *  video handler to be called ASAP.
 */
static void
interrupt_handler(void *v)
{
	struct videum_device *dev = (struct videum_device *)v;

	if (!dev->ints_enabled ||
	     dev->dma_state == DMA_STATE_ARMING)
	{
		return;
	}

	/*  Call "bottom half" of handler  */
	dev->tqnode_dpc.routine = dev->video_handler;
	dev->tqnode_dpc.data = dev;
	queue_task(&dev->tqnode_dpc, &tq_immediate);
	mark_bh(IMMEDIATE_BH);

	if ((dev->irq == 0 || dev->dma_mode == DMAMODE_SLAVE) &&
	    dev->ints_enabled)
	{/*	Poll again on next timer tick  */
		dev->tlnode.expires = jiffies + HZ/100;
		add_timer(&dev->tlnode);
	}
}

/*  Called by the system when our hardware interrupt occurs
 */
static void
interrupt_hw(int irq, void *v,
	     struct pt_regs *regs)
{
	struct videum_device	*dev = (struct videum_device *)v;
	int			handled = 0;

	if (v == NULL)
		return;
	if (dev->type == VIDEUM_TYPE_PCI)
	{
		__u8	t;

		BEGIN_CRITICAL_SECTION(dev);
		t = inb(BMSTATUS1(dev));
		if (!(t & BMSTATUS_DMA_INT))
		{/* Not us. Likely another device is sharing the IRQ */
			//debug_msg("HW interrupt (not Videum)\n");
			handled = 1;
		}
		else
		{
			outb(t, BMSTATUS1(dev));
			t = inb(DMAC(dev));
			t |=  DMAC_HIRQREQn;
			outb(t, DMAC(dev));
			if (dev->dma_state == DMA_STATE_ARMING)
			{
				//debug_msg("HW interrupt (arming)\n");
				dev->dma_state = DMA_STATE_RUNNING;
				handled = 1;
			}
		}
		END_CRITICAL_SECTION(dev);
	}
	if (handled)
		return;

	//debug_msg("HW interrupt (real)\n");
	interrupt_handler(v);
}

void
wnv_interrupt_disable(struct videum_device *dev)
{
	if (!dev->ints_enabled)
		return;
	dev->ints_enabled = 0;
	wavi_writereg(dev, WAVI_ISR, 0);
	if (dev->irq == 0 || dev->dma_mode == DMAMODE_SLAVE)
	{
		del_timer(&dev->tlnode);
	}
	if (dev->type == VIDEUM_TYPE_PCI)
	{
		__u8	t;
		t = inb(DMAC(dev));
		t &= ~DMAC_HIRQEN;
		outb(t, DMAC(dev));
	}
	wake_up_interruptible(&dev->new_video_frame);
}

void
wnv_interrupt_enable(struct videum_device *dev)
{
	if (dev->video_handler == NULL)
		return;
	if (dev->ints_enabled)
		wnv_interrupt_disable(dev);
	dev->ints_enabled = 1;

	if (dev->type == VIDEUM_TYPE_PCI)
	{
		__u8	t;
		t = inb(DMAC(dev));
		t |=  DMAC_HIRQEN;
		t |=  DMAC_HIRQREQn;/* disable interrupt request bit */
		outb(t, DMAC(dev));
	}
	wavi_readreg(dev, WAVI_ITR);

	wavi_writereg(dev, WAVI_ISR, WAVI_VIEBIT);

	dev->tlnode.function = NULL;
	if (dev->irq == 0 || dev->dma_mode == DMAMODE_SLAVE)
	{
		init_timer(&dev->tlnode);
		dev->tlnode.function =
			(void(*)(unsigned long))interrupt_handler;
		dev->tlnode.data = (unsigned long)dev;
		dev->tlnode.expires = jiffies + HZ/100;
		add_timer(&dev->tlnode);
	}
//debug_msg("%s interrupts enabled\n",
//	  dev->tlnode.function?"Polled":"Hardware");
}

int
wnv_register_video_interrupt(struct videum_device *dev,
			     void (*handler)(void *))
{
	if (dev->pci && dev->irq == 0 && !wnv_noirq)
	{
		dev->irq = dev->pci->irq;
		if (request_irq(dev->irq, interrupt_hw, SA_SHIRQ,
				dev->shortname, dev) < 0)
		{
			err_msg("Denied IRQ %d\n", dev->irq);
			dev->irq = 0;
		}
	}
	dev->video_handler = handler;
	return 0;
}

void
wnv_unregister_video_interrupt(struct videum_device *dev)
{
	wnv_interrupt_disable(dev);
	if (dev->irq)
		free_irq(dev->irq, dev);
	dev->irq = 0;
}


/*  =====================================================================
 *	The functions below this point are only called during loading
 *	and unloading of the driver.
 */

/*
 *
 *	L O W - L E V E L   P n P   R O U T I N E S
 *
 */

static PNP_INFO	glb_pnp_info[PNP_MAX_BOARDS];
static u8	ReadPort	= 0x80;  //default.            
static int	LastCSN		= 0;
static int	ReadPortFrozen	= 0;                                       
static int	MaxBoardIndex	= 0;

static void
pnp_writereg(u8 RegIndex, u8 Data)
{                                
	outb(RegIndex, ADDRESS_PORT);
	outb(Data, WRITE_DATA_PORT);
}

static u8
pnp_readreg(u8 RegIndex)
{
	outb(RegIndex, ADDRESS_PORT);
	return inb(READPORT(ReadPort));
}

static int
pnp_putinwaitforkeystate(void)
{
	pnp_writereg(CONFIG_CONTROL, GOTO_WAIT_FOR_KEY);
	return 1;
}

static int
pnp_writekey(void)
{                            
	pnp_putinwaitforkeystate(); //in case we are not in wait for key state!
	//put boards in configuration mode...
	/* Resetting LFSR */
	outb(0x00, ADDRESS_PORT);	outb(0x00, ADDRESS_PORT);
	/* Following 32 writes form the LFSR sequence */
	outb(0x6a, ADDRESS_PORT);	outb(0xb5, ADDRESS_PORT);
	outb(0xda, ADDRESS_PORT);	outb(0xed, ADDRESS_PORT);
	outb(0xf6, ADDRESS_PORT);	outb(0xfb, ADDRESS_PORT);
	outb(0x7d, ADDRESS_PORT);	outb(0xbe, ADDRESS_PORT);
	outb(0xdf, ADDRESS_PORT);	outb(0x6f, ADDRESS_PORT);
	outb(0x37, ADDRESS_PORT);	outb(0x1b, ADDRESS_PORT);
	outb(0x0d, ADDRESS_PORT);	outb(0x86, ADDRESS_PORT);
	outb(0xc3, ADDRESS_PORT);	outb(0x61, ADDRESS_PORT);
	outb(0xb0, ADDRESS_PORT);	outb(0x58, ADDRESS_PORT);
	outb(0x2c, ADDRESS_PORT);	outb(0x16, ADDRESS_PORT);
	outb(0x8b, ADDRESS_PORT);	outb(0x45, ADDRESS_PORT);
	outb(0xa2, ADDRESS_PORT);	outb(0xd1, ADDRESS_PORT);
	outb(0xe8, ADDRESS_PORT);	outb(0x74, ADDRESS_PORT);
	outb(0x3a, ADDRESS_PORT);	outb(0x9d, ADDRESS_PORT);
	outb(0xce, ADDRESS_PORT);	outb(0xe7, ADDRESS_PORT);
	outb(0x73, ADDRESS_PORT);	outb(0x39, ADDRESS_PORT);
	return 1;
}

static int
pnp_findwinnovdevices(void)
{
	int		i, j, BufLen;
	unsigned int	checksum	= 0;
	u8		Prom[1100];
	char		Buff[5];
	u32		dwBoardFound	= 0;
	u32		dwDevCount	= 0;
	int		freturn		= 0;

	if (LastCSN == 0)
		return 0;                                           
	pnp_writekey();
	for (i = 1; i <= LastCSN; i++) 
	{
		dwBoardFound = 0;
		checksum = 0;    
		dwDevCount = 0;
		BufLen = 0;
		pnp_writereg(WAKE, (u8)i); // select a board
		j = 0;
		while ((j<1023) && (Prom[j-1]!=0x79))  //look for end tag...
		{
			while ((pnp_readreg(STATUS) & 0x01)== 0)
				continue; // Wait until next byte is ready!
			Prom[j] = pnp_readreg(RESOURCE_DATA);
			if (j > 8)
				checksum += Prom[j++];
			else
				j++;
		}
		while ((pnp_readreg(STATUS) & 0x01) == 0)
			continue; // Wait until next byte is ready!
		Prom[j] = pnp_readreg(RESOURCE_DATA);  //read checksum!
		checksum = ~checksum;
		checksum++;
		BufLen = j - 1;  //	data length	
		if (Prom[j] != (unsigned char)checksum)  //wrong checksum!
		{
			debug_msg("PnP checksum error\n");
			pnp_putinwaitforkeystate();
			return freturn;
		}

		j = 9; //start of resources...
		while (j <= BufLen)
		{
			if (dwBoardFound > 0 &&//looking for I/O resources...
			    ((Prom[j] & 0x80) == 0) &&
			    (((Prom[j] & 0x78)>>3) == 0x8)) //I/O descriptor!
			{
				glb_pnp_info[MaxBoardIndex].dwDevIndex = 
					dwDevCount-1;
				sprintf(glb_pnp_info[MaxBoardIndex].ID,
					"WNV%04x",(unsigned int)dwBoardFound);
				glb_pnp_info[MaxBoardIndex].wStartAddress =
					Prom[j+2] + (256 * Prom[j+3]);
				glb_pnp_info[MaxBoardIndex].wEndAddress =
					Prom[j+4] + (256 * Prom[j+5]);
				glb_pnp_info[MaxBoardIndex].wLength =
					Prom[j+7];
				glb_pnp_info[MaxBoardIndex].wIncrement =
					Prom[j+6];
				glb_pnp_info[MaxBoardIndex++].dwHWIndex =
					(u32)i;
				dwBoardFound = 0;
				freturn = 1;
			}
			if (((Prom[j] & 0x80) == 0) &&
			    (((Prom[j] & 0x78) >> 3) == 0x2)) //device ID!
			{                               
				dwDevCount++;
				Buff[0] = (((Prom[j+1] & 0x7F) >> 2) + 0x40);
				Buff[1] = ((((Prom[j+1] & 0x03) << 3)
					  | (Prom[j+2] >> 5)) + 0x40);
				Buff[2] = ((Prom[j+2] & 0x1F) + 0x40);
				Buff[3] = 0;
				dwBoardFound = 0; // might not be ours ...
				if (strcmp(Buff, WINNOV_MFG_ID) == 0 &&
				    MaxBoardIndex < PNP_MAX_BOARDS)
				{
					dwBoardFound = (u32)
						(Prom[j+4]+(256 * Prom[j+3]));
					//does it support rangecheck?
					glb_pnp_info[MaxBoardIndex].
						fRangeCheck =
						(((Prom[j+5] & 0x2)!=0) ? 
						 1:0);
				}
			}
			if ((Prom[j] & 0x80) == 0)   //small resource data!
				j += (Prom[j] & 0x7) + 1;
			else
				j += Prom[j+1] + (256 * Prom[j+2]) + 3;
		}
	}
	pnp_putinwaitforkeystate();
	return freturn;
}

static int
pnp_isolatehw(void)
{
static	int	already_done	= 0;
	int	step		= 0;
	u8	checksum	= 0x6A;  //lfsr initial value!
	u8	read1, read2;
	int	detected	= 0;
	int	LastBit		= 0;
	int	Fail		= 0;
	int	freturn;
		
	if (already_done)
		return 1;
	already_done = 1;

	//isolate all PNP boards and assign CSNs 
	pnp_writekey();  //prepare boards for configuration
	//Clear all CSN so that we can enumerate from 1
	pnp_writereg(CONFIG_CONTROL, RESET_CSN);
	LastCSN = 0;
	MaxBoardIndex = 0;
	// Sets the address of the READ_DATA port...
	pnp_writereg(RD_DATA, ReadPort);
	while (!Fail)
	{
		// put all boards (with CSN!=0) into isolation state
		pnp_writereg(WAKE, 0);
		udelay(1000);
		while(step < 64)// Get 64 first bits of the Next isolated card!
		{
			udelay(250);
			read1 = pnp_readreg(SERIAL_ISOLATION);
			read2 = pnp_readreg(SERIAL_ISOLATION);
			LastBit = 0;
			if ((read1 == 0x55) && (read2 == 0xAA))
			{
				LastBit  = 1;
				detected = 1;
			}
			if (((checksum & 0x01) + ((checksum >> 1) & 0x01) 
			      + ((LastBit) ? 1:0) ) & 0x01)
				checksum = 0x80 | (checksum >> 1);
			else
				checksum = (checksum >> 1);
			step++;
		}   
		while ((step < 72) && (!Fail))
		{
			udelay(250);
			read1 = pnp_readreg(SERIAL_ISOLATION);
			read2 = pnp_readreg(SERIAL_ISOLATION);
			LastBit = 0;
			if ((read1 == 0x55) && (read2 == 0xAA))
			{
				LastBit=1;
				detected=1;
			}
			if (((!LastBit) && (checksum & 0x01)) 
			     || (LastBit && (!(checksum & 0x01))) )
			{
				Fail = 1;
			}
			checksum = (checksum >> 1);
			step++;
		}
		if (!ReadPortFrozen && (Fail || !detected))
		{
			Fail = 1;
			if (ReadPort < MAX_READPORT)
			{
				ReadPort++;
				// Sets the address of the READ_DATA port...
				pnp_writereg(RD_DATA, ReadPort);
				// not a real failure... may be just a 
                                // conflict on READ_DATA port...
				Fail = 0;
				// make sure we sieve through the other cases!
				detected = 0;
				udelay(1000); //need to wait 1ms
			}
		}
		else if (!detected)
			Fail = 1;
		if ((!Fail) && (detected))
		{
			ReadPortFrozen = 1;
			//assign CSN to board!
			pnp_writereg(CSN, (u8)++LastCSN);
		}
		step = 0;
		checksum = 0x6A;
		detected = 0;
		LastBit = 0;
	}
	pnp_putinwaitforkeystate();
	if (LastCSN == 0)
		return 0;                           
	freturn = pnp_findwinnovdevices();
	return freturn;
}

static int
pnp_assignresources(u32		dwIndex,
		    RES_INFO	*resinfo)
{
	int	freturn = 0;
	u16	port;
	int	i;

	pnp_writekey();
	pnp_writereg(WAKE, (u8) glb_pnp_info[dwIndex].dwHWIndex);
	pnp_writereg(LOGICAL_DEVICE_NUMBER,
		     (u8)glb_pnp_info[dwIndex].dwDevIndex); //device number
	pnp_writereg(ACTIVATE,0); // deactivate
	pnp_writereg(IO_RANGE_CHECK,0); //disable range check for now!
	//ready to program resources!

	//memory descriptor 0...3
	for (i = 0; i < 4; ++i)
	{/* Should be passed in RES_INFO, but not needed for Videum */
		pnp_writereg(MEM_BASEH   + i * MEM_DESC_INCREMENT, 0);
		pnp_writereg(MEM_BASEL   + i * MEM_DESC_INCREMENT, 0);
		pnp_writereg(MEM_CTL     + i * MEM_DESC_INCREMENT, 0);
		pnp_writereg(MEM_LENGTHH + i * MEM_DESC_INCREMENT, 0);
		pnp_writereg(MEM_LENGTHL + i * MEM_DESC_INCREMENT, 0);
	}
	//32 bit memory space configuration: descriptor 0...3
	for (i = 0; i < 4; ++i)
	{/* Should be passed in RES_INFO, but not needed for Videum */
		pnp_writereg(MEM32_BASEHH   + i * MEM32_DESC_INCREMENT, 0);
		pnp_writereg(MEM32_BASEHL   + i * MEM32_DESC_INCREMENT, 0);
		pnp_writereg(MEM32_BASELH   + i * MEM32_DESC_INCREMENT, 0);
		pnp_writereg(MEM32_BASELL   + i * MEM32_DESC_INCREMENT, 0);
		pnp_writereg(MEM32_CTL      + i * MEM32_DESC_INCREMENT, 0);
		pnp_writereg(MEM32_LENGTHHH + i * MEM32_DESC_INCREMENT, 0);
		pnp_writereg(MEM32_LENGTHHL + i * MEM32_DESC_INCREMENT, 0);
		pnp_writereg(MEM32_LENGTHLH + i * MEM32_DESC_INCREMENT, 0);
		pnp_writereg(MEM32_LENGTHLL + i * MEM32_DESC_INCREMENT, 0);
	}
	//I/O configuration registers : descriptor 0...7
	for (i = 0; i < 8; ++i)
	{
		pnp_writereg(IO_BASEH + i * IO_DESC_INCREMENT,
			     (u8)(resinfo->wBasePort[i] >> 8));
		pnp_writereg(IO_BASEL + i * IO_DESC_INCREMENT,
			     (u8)(resinfo->wBasePort[i]));
	}
	//Interrupt descriptor 0
	pnp_writereg(INT_LEVEL,0);
	pnp_writereg(INT_TYPE,0);
	//Interrupt descriptor 1
	pnp_writereg(INT_LEVEL+INT_DESC_INCREMENT,0);
	pnp_writereg(INT_TYPE+INT_DESC_INCREMENT,0);
	//DMA descriptor 0
	pnp_writereg(DMA_SELECT_0, 0x04);
	//DMA descriptor 1
	pnp_writereg(DMA_SELECT_1, 0x04);
	// end of program ressources...

	// test for conflicts!
	if (glb_pnp_info[dwIndex].fRangeCheck == 1)
	{
		//enable range check and return 0xAA	
		pnp_writereg(IO_RANGE_CHECK, 0x2);
		for(port = resinfo->wBasePort[0];
		    (port < (resinfo->wBasePort[0] + resinfo->wLength[0])
		     && (!freturn));
		    port++)
		{
			if (inb(port) != 0xAA)
				freturn = 1;
		}
		if (!freturn)
		{
			//enable range check and return 0x55
			pnp_writereg(IO_RANGE_CHECK, 0x3);
			for(port=resinfo->wBasePort[0];
			    (port<(resinfo->wBasePort[0]+resinfo->wLength[0])&&
			     (!freturn));
			    port++)
			{
				if (inb(port) != 0x55)
					freturn=1;
			}
		}
		pnp_writereg(IO_RANGE_CHECK, 0); //disable range check!
	}
	if (!freturn)
		pnp_writereg(ACTIVATE,0x1); //activate Board!
	pnp_putinwaitforkeystate();
	return !freturn;
}


/*
 *	D E V I C E   I N I A L I Z A T I O N   R O U T I N E S
 *
 *	These routines locate and enable the Videum, and initialize
 *	the device structure. After enabling, the WAVI register access
 *	functions and I2C functions will work.
 */

#ifndef PCI_DEVICE_ID_CMD_647
#define PCI_DEVICE_ID_CMD_647	0x0647
#endif

static int
config_pci_videum(struct videum_device *dev,
		  int idev)
{
#ifndef CONFIG_PCI

/*	Kernel is not configured for PCI bus support	*/
	return 0;

#else

	static struct pci_dev	*lastpci	= NULL;
	__u8	al;
	__u32	dw;

	if (idev > 3 || !pci_present())
		return 0;

	dev->pci = pci_find_device(PCI_VENDOR_ID_CMD, 
				   PCI_DEVICE_ID_CMD_647, lastpci);
	if (dev->pci == NULL)
		return 0;/*  No more Videums  */
	lastpci = dev->pci;

	/*  Found a device with the right ID, verify it's a Videum.	*/
	/*  Base class has to be MULTIMEDIA	*/
	if ((lastpci->class >> 16) != PCI_BASE_CLASS_MULTIMEDIA)
		return 0;/*  Not a Videum  */

	/*  It's most likely a Videum. Keep this pci_dev pointer	*/
	dev->pci = lastpci;

	/*  Five I/O ranges	*/
	if (check_region(PCIBA0(dev),  8) ||
	    check_region(PCIBA1(dev),  4) ||
	    check_region(PCIBA2(dev),  8) ||
	    check_region(PCIBA3(dev),  4) ||
	    check_region(PCIBA4(dev), 16))
	{
		info_msg("Possible IO conflict with PCI Videum\n");
	}
	request_region(PCIBA0(dev),  8, dev->shortname);
	request_region(PCIBA1(dev),  4, dev->shortname);
	request_region(PCIBA2(dev),  8, dev->shortname);
	request_region(PCIBA3(dev),  4, dev->shortname);
	request_region(PCIBA4(dev), 16, dev->shortname);

	/*  Turn on the device	*/
	pci_read_config_dword(dev->pci, PCI_COMMAND, &dw);
	dw |=  PCI_COMMAND_IO;  /* chip enable -- needed to talk to board...*/
	dw |=  PCI_COMMAND_MASTER;  /* bus master enable  */
	dw &= ~PCI_COMMAND_PARITY;  /* ignore parity  */
	dw &= ~(PCI_STATUS_PARITY << 16);
	dw &= ~(PCI_STATUS_REC_TARGET_ABORT << 16);
	dw &= ~(PCI_STATUS_REC_MASTER_ABORT << 16);
	dw &= ~(PCI_STATUS_DETECTED_PARITY << 16);
	pci_write_config_dword(dev->pci, PCI_COMMAND, dw);

	/*  Initialize the PCI IDE controller	*/

	pci_read_config_byte(dev->pci, CMD_CONFIG_CNTRL, &al);
	al |=  (1<< 3); /* second channel control enable */
	al |=  (1<< 6); /* disable drive 0 read ahead */
	al |=  (1<< 7); /* disable drive 1 read ahead */
	pci_write_config_byte(dev->pci, CMD_CONFIG_CNTRL, al);

	al = 0x21;  /* R/W active 2 clocks-- R/W recovery 2 Clocks-- */
	pci_write_config_byte(dev->pci, CMD_CONFIG_CMDTIM, al);

	pci_read_config_byte(dev->pci, CMD_CONFIG_ARTTIM23, &al);
	al |=  (1<< 2); /* disable drive 2 read ahead */
	al |=  (1<< 3); /* disable drive 3 read ahead */
	al |=  (1<< 6); /* address setup time 2 clocks */
	al &= ~(1<< 7); /* address setup time 2 clocks */
	pci_write_config_byte(dev->pci, CMD_CONFIG_ARTTIM23, al);

	pci_read_config_byte(dev->pci, CMD_CONFIG_ARTTIM0, &al);
	al |=  (1<< 6); /* address setup time 2 clocks */
	al &= ~(1<< 7); /* address setup time 2 clocks */
	pci_write_config_byte(dev->pci, CMD_CONFIG_ARTTIM0, al);

	pci_read_config_byte(dev->pci, CMD_CONFIG_ARTTIM1, &al);
	al |=  (1<< 6); /* address setup time 2 clocks */
	al &= ~(1<< 7); /* address setup time 2 clocks */
	pci_write_config_byte(dev->pci, CMD_CONFIG_ARTTIM1, al);

	al = 0x21;  /* R/W active 2 clocks-- R/W recovery 2 Clocks-- */
	pci_write_config_byte(dev->pci, CMD_CONFIG_DRWTIM0, al);
	pci_write_config_byte(dev->pci, CMD_CONFIG_DRWTIM1, al);
	pci_write_config_byte(dev->pci, CMD_CONFIG_DRWTIM2, al);
	pci_write_config_byte(dev->pci, CMD_CONFIG_DRWTIM3, al);

	al = inb(BMCMD0(dev));
	al &= ~BMCMD_START_STOP; /* disengage bus master channel 0 */
	al |=  BMCMD_RW_CNRL;    /* PCI bus master writes are performed */
	outb(al, BMCMD0(dev));
	outb(0, MRDMODE(dev));   /* memory read (default) */
	al = inb(BMSTATUS0(dev));
	al |=  (1<< 5); /* dma OK on drive 2 */
	al |=  (1<< 6); /* dma OK on drive 1 */
	outb(al, BMSTATUS0(dev));

	al = inb(BMCMD1(dev));
	al &= ~BMCMD_START_STOP; /* disengage bus master channel 1 */
	al |=  BMCMD_RW_CNRL;    /* PCI bus master writes are performed */
	outb(al, BMCMD1(dev));
	al = inb(BMSTATUS1(dev));
	al |=  (1<< 5); /* dma OK on drive 4 */
	al |=  (1<< 6); /* dma OK on drive 3 */
	outb(al, BMSTATUS1(dev));

	/*  Init Videum IDE  */
	al = inb(DMAC(dev));
	al &= ~DMAC_HIRQEN;   /* Interrupts disabled */
	al |=  DMAC_HDMAEN1;  /* DMA state machine stopped */
	al |=  DMAC_RSTMDMA1; /* reset MDMAEN1 flop */
	al &= ~DMAC_SWAP;     /* Disable byte swapper */
	al |=  DMAC_HIRQREQn; /* No interrupt request */
	outb(al, DMAC(dev));
	al &= ~DMAC_RSTMDMA1; /* release Reset */
	outb(al, DMAC(dev));
	dev->dma_state = DMA_STATE_IDLE;

	info_msg("Found PCI Videum (slot %d)\n",
		 PCI_SLOT(dev->pci->devfn));
	dev->type = VIDEUM_TYPE_PCI;
	dev->sram_base = 0x60000;
	dev->sram_size = 0x40000;
	dev->irq = 0;/* we will grab it later */
	dev->index = idev;

	return 1;
#endif
}

static void
unconfig_pci_videum(struct videum_device *dev)
{
#ifndef CONFIG_PCI

/*	Kernel is not configured for PCI bus support	*/
	return;

#else
	__u32		dw;

	if (dev == NULL || dev->pci == NULL ||
	    dev->type != VIDEUM_TYPE_PCI)
		return;
	if (!pci_present())
		return;/*  No PCI bus  */

	/*  Turn off the device	*/
	pci_read_config_dword(dev->pci, PCI_COMMAND, &dw);
	dw &= ~PCI_COMMAND_IO;  /* chip enable -- needed to talk to board...*/
	dw &= ~PCI_COMMAND_MASTER;  /* bus master disable  */
	dw &= ~PCI_COMMAND_PARITY;  /* ignore parity  */
	pci_write_config_dword(dev->pci, PCI_COMMAND, dw);

	release_region(PCIBA0(dev),  8);
	release_region(PCIBA1(dev),  4);
	release_region(PCIBA2(dev),  8);
	release_region(PCIBA3(dev),  4);
	release_region(PCIBA4(dev), 16);
	if (dev->irq)
		free_irq(dev->irq, dev);

	debug_msg("Turned off PCI Videum\n");
#endif
}

static int
config_isa_videum(struct videum_device *dev,
		  int idev)
{
	int	x;
	int	set;
	int	base;
	int	forced_base = 0;
	int	io;
	int	err	= -1;

	if (idev == 0)
	{
		dev->cfg = WAVI_BOARD0CONFIG;
		forced_base = isa0_iobase;
	}
	else if (idev == 1)
	{
		dev->cfg = WAVI_BOARD1CONFIG;
		forced_base = isa1_iobase;
	}
	else
		return 0;
	dev->index = idev;

	dev->type = VIDEUM_TYPE_ISA;
	for (set = 1; set < 4; ++set)
	{
		base = 0x170 - idev * 0x040 - set * 0x010;
		if (forced_base && base != (forced_base & 0x7FFF))
			continue;
		for (io = base, err = 0; io < 0xFFFF; io += 0x0400)
		{
			if ((err = check_region(io, 4)) < 0)
				break;
		}
		if (err < 0 && base != (forced_base & 0x7FFF))
			continue;
		dev->iobase = forced_base ? forced_base : base;
		outb(set, dev->cfg);

		/* BUG: This probe could be more thorough */
		x = wavi_readreg(dev, WAVI_MMA) & WAVI_FULLID;
		if (x == 0x02 || x == 0x12 || x == 0x22)
			break;
		err = -1;
		if (base == (forced_base & 0x7FFF))
			break;
	}
	if (err < 0 && base != (forced_base & 0x7FFF))
	{
		dev->type = 0;
		dev->iobase = 0;
		return 0;
	}

	for (io = base; io < 0xFFFF; io += 0x0400)
		request_region(io, 4, dev->shortname);
	request_region(dev->cfg, 1, dev->shortname);

	info_msg("Found legacy ISA Videum (#%d, %03Xh)\n", 
		 dev->index, dev->iobase);
	if (dev->iobase == forced_base)
		info_msg("(IO port forced by command line argument)\n");
	if (err < 0)
		info_msg("Possible IO conflict detected\n");
	dev->type = VIDEUM_TYPE_ISA;
	dev->sram_base = 0x00000;
	dev->sram_size = 0x10000;
	dev->irq = 0;/* We will use the jiffy timer interrupt to poll */
	return 1;
}

static void
unconfig_isa_videum(struct videum_device *dev)
{
	int	io;

	outb(WAVI_PORTDESELECT, dev->cfg);
	if (dev->iobase)
	{
		for (io = (dev->iobase & 0x7FFF); io < 0xFFFF; io += 0x0400)
			release_region(io, 4);
		release_region(dev->cfg, 1);
	}
	debug_msg("Turned off ISA Videum (#%d)\n", dev->index);
}

static int
config_pnpisa_videum(struct videum_device *dev,
		     int idev)
{
	static int	ipnp	= 0;
	RES_INFO	resinfo;
	int		err	= -1;
	int		io;

	if (idev > 0)
		return 0;/* There can only be one ISAPnP Videum */

	/*  Enumerate the cards, and hunt for a Videum in the list */
	if (!pnp_isolatehw())
		return 0;
	while (ipnp < PNP_MAX_BOARDS &&
	       strcmp(glb_pnp_info[ipnp].ID, WINNOV_PNP_VIDEUM1_ID) != 0)
		++ipnp;
	if (ipnp >= PNP_MAX_BOARDS)
		return 0;/* no more Videums */

	/*  Find an available IO range and enable the card there */
	memset(&resinfo, 0, sizeof(resinfo));
	resinfo.wLength[0] = glb_pnp_info[ipnp].wLength;
	for (io = glb_pnp_info[ipnp].wStartAddress;
	     io + resinfo.wLength[0] < glb_pnp_info[ipnp].wEndAddress;
	     io += glb_pnp_info[ipnp].wIncrement)
	{
		if (pnp0_iobase && io != pnp0_iobase)
			continue;
		err = check_region(io, resinfo.wLength[0]);
		if (err < 0 && io != pnp0_iobase)
			continue;
		resinfo.wBasePort[0] = io;
		if (pnp_assignresources(0, &resinfo))
			break;
		err = -1;
		if (io == pnp0_iobase)
			break;
	}
	if (err < 0 && resinfo.wBasePort[0] != pnp0_iobase)
	{
		dev->type = 0;
		dev->iobase = 0;
		return 0;
	}
	dev->iobase = resinfo.wBasePort[0];
	request_region(dev->iobase, resinfo.wLength[0], dev->shortname);

	info_msg("Found PnP ISA Videum (%04Xh)\n", dev->iobase);
	if (dev->iobase == pnp0_iobase)
		info_msg("(IO port forced by command line argument)\n");
	if (err < 0)
		info_msg("Possible IO conflict detected\n");
	dev->type = VIDEUM_TYPE_PNPISA;
	dev->sram_base = 0x00000;
	dev->sram_size = 0x10000;
	dev->irq = 0;/* We will use the jiffy timer interrupt to poll */
	dev->index = idev;
	return 1;
}

static void
unconfig_pnpisa_videum(struct videum_device *dev)
{
	if (dev->iobase)
		release_region(dev->iobase, 0x80);
	/* BUG: disable the device */
	debug_msg("Turned off PnP ISA Videum.\n");
}

static void
unconfig_any_videum(struct videum_device *dev)
{
	wnv_interrupt_disable(dev);
	switch (dev->type)
	{
	case VIDEUM_TYPE_PCI:
		unconfig_pci_videum(dev);
		break;
	case VIDEUM_TYPE_ISA:
		unconfig_isa_videum(dev);
		break;
	case VIDEUM_TYPE_PNPISA:
		unconfig_pnpisa_videum(dev);
		break;
	}
	memset(dev, 0, sizeof(videum[0]));
}

static int
config_any_videum(struct videum_device *dev)
{
	static int	pcidev	  = 0;
	static int	isadev	  = 0;
	static int	pnpisadev = 0;

	sprintf(dev->shortname, "videum%d", dev - videum);
	dev->lock = SPIN_LOCK_UNLOCKED;

	if (!config_pci_videum(dev, pcidev++) &&
	    !config_isa_videum(dev, isadev++) &&
	    !config_pnpisa_videum(dev, pnpisadev++))
		return 0;

	/*  The WAVI registers are now accessible!	*/
	wavi_initialize(dev);

	if (!eeprom_load(dev))
	{
		err_msg("EEPROM read failure\n");
		unconfig_any_videum(dev);
		return 0;
	}

	/*  SRAM memory map	*/
	if (dev->eeprom.bBoardType == TYPE_VIDEUM)
		dev->sram_base = 0x10000;/* correction for early Videums */
	dev->sram_audio		= dev->sram_base;
	if (dev->wavi_type == WINNOV_WAVI_93 ||     /* wnvaud doesn't work */
	    dev->wavi_type == WINNOV_WAVI_95 ||     /* with WAVI 93 or 95  */
	    dev->eeprom.bMinorType == MINORTYPE_VO) /* not Video-Only      */
		dev->sram_audio_size = 0;
	else
		dev->sram_audio_size = WAVI_AUDBUFSIZE;
	dev->sram_ucode		= dev->sram_audio + dev->sram_audio_size;
	dev->sram_ucode_size	= UC_SIZE;
	dev->sram_linebuf	= dev->sram_ucode + dev->sram_ucode_size;
	dev->sram_linebuf_size	= 0x00800;
	dev->sram_video		= dev->sram_linebuf + dev->sram_linebuf_size;
	dev->sram_video_size	= dev->sram_base + dev->sram_size 
		                - dev->sram_video;

	/*  Descriptive device name  */
	sprintf(dev->v.name, "%s %s (%d)", dev->eeprom.szOemName,
		dev->eeprom.szProduct, dev - videum);
	return 1;
}

/*
 *	M O D U L E   I N I T   A N D   C L E A N U P
 */

int
init_module(void)
{
	int	i;

	for (i = 0; i < WNV_NBOARDS; ++i)
	{
		memset(&videum[i], 0, sizeof(videum[0]));
		if (!config_any_videum(&videum[i]))
		{
			break;
		}
	}
	if (i == 0)
	{
		err_msg("Videum not found.\n");
		return -ENODEV;/* cleanup will not be called */
	}
	return 0;
}

void
cleanup_module(void)
{
	int	i;

	for (i = 0; i < WNV_NBOARDS; ++i)
		unconfig_any_videum(&videum[i]);
}

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

/*
 *	Exports
 */

EXPORT_SYMBOL(videum);
EXPORT_SYMBOL(wavi_readreg);
EXPORT_SYMBOL(wavi_writereg);
EXPORT_SYMBOL(wavi_block_read);
EXPORT_SYMBOL(wavi_block_write);
EXPORT_SYMBOL(wavi_benchmark);
EXPORT_SYMBOL(wnv_bm_cleanup);
EXPORT_SYMBOL(wnv_bm_resync);
EXPORT_SYMBOL(wnv_bm_disarm);
EXPORT_SYMBOL(wnv_bm_arm);
EXPORT_SYMBOL(wnv_bm_is_done);
EXPORT_SYMBOL(wnv_bm_build_scatter_list);
EXPORT_SYMBOL(wnv_i2c_get_ready);
EXPORT_SYMBOL(wnv_i2c_read_reg_byte);
EXPORT_SYMBOL(wnv_i2c_write_reg_byte);
EXPORT_SYMBOL(wnv_i2c_read_reg_word);
EXPORT_SYMBOL(wnv_i2c_write_reg_word);
EXPORT_SYMBOL(wnv_register_video_interrupt);
EXPORT_SYMBOL(wnv_unregister_video_interrupt);
EXPORT_SYMBOL(wnv_interrupt_enable);
EXPORT_SYMBOL(wnv_interrupt_disable);
